/*
  File name        : childcontrol.c
  Date of creation : 15/11/2008
  Version          : 1.0
  Copyright        : Soft'n'design
  Author           : Laurent Mazet <mazet@softndesign.org>

  Description      : This file contains mechanism to control child processes

  History          :
  - initial version
*/

/* depend: */
/* cflags: */
/* lflags: */

#include <assert.h>
#include <errno.h>
#include <malloc.h>
#include <pthread.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/ptrace.h>
#include <sys/resource.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>

#include "verbose.h"

#include "childcontrol.h"

/**
   Return codes
*/
#define CHILD_CONTROL_RC_NORMAL_END 0
#define CHILD_CONTROL_RC_ASSERT_END 1
#define CHILD_CONTROL_RC_MAX_DEAD_END 2

/**
   Default maximun number of dead children
*/
#define CHILD_CONTROL_MAX_DEAD 10

/**
   Verbose level
*/
DECLARE_VERBOSE_LEVEL (child_control, 1);

/**
   Child structure
*/
typedef struct {

    /* Child pid */
    pid_t pid;

    /* Child id */
    int id;

    /* Generic parameter */
    void *data;

    /* Prefork init function */
    int (*preinit) (void *data);

    /* Postfork init function */
    int (*postinit) (void *data);

    /* Child main process */
    int (*process) (void *data);

    /* Child termination on signal function */
    int (*term) (void *data);

    /* Number of dead */
    int number_of_dead;

} child_t;

/**
   Children control structure
*/
typedef struct 
{
    /* Father pid; if it contains 0, this means that child control has
     * not been initialised properly */
    pid_t father_pid;

    /* Monitor posix thread */
    pthread_t dispatcher_thread;

    /* Child pid */
    pid_t child_pid;

    /* Number of children */
    int number_of_children;

    /* List of children */
    child_t **list_of_children;

    /* Number of dead */
    int number_of_dead;

    /* Max number of dead */
    int max_number_of_dead;

    /* Killing */
    int killing;

} child_control_t;

/**
   Children control. This variable is set at init and contains all
   needed variables
*/
static child_control_t *_child_control = NULL;

static child_t *_child = NULL;

/**
   Get resouce usage

   @param who type of resource
*/
void get_resource_usage (void)
{

    /* get usage statistics */
    VERBOSE (child_control, 2, printf ("getting resource usage\n"));
    struct rusage usage;
    int rc = getrusage (RUSAGE_SELF, &usage);
    if (rc) {
        printf ("error getrusage (%d), %s\n", errno, strerror (errno));
    } else {
        printf ("self: utime=%0lu.%06lu stime=%0lu.%06lu maxrss=%ld ixrss=%ld idrss=%ld isrss=%ld pf=%ld vcs=%ld ivcs=%ld\n",
                usage.ru_utime.tv_sec, usage.ru_utime.tv_usec, usage.ru_stime.tv_sec, usage.ru_stime.tv_usec, usage.ru_maxrss,
                usage.ru_ixrss, usage.ru_idrss, usage.ru_isrss, usage.ru_majflt, usage.ru_nvcsw, usage.ru_nivcsw);
    }

    rc = getrusage (RUSAGE_CHILDREN, &usage);
    if (rc) {
        printf ("error getrusage (%d), %s\n", errno, strerror (errno));
    } else {
        printf ("children: utime=%lu.%06lu stime=%lu.%06lu maxrss=%ld ixrss=%ld idrss=%ld isrss=%ld pf=%ld vcs=%ld ivcs=%ld\n",
                usage.ru_utime.tv_sec, usage.ru_utime.tv_usec, usage.ru_stime.tv_sec, usage.ru_stime.tv_usec, usage.ru_maxrss,
                usage.ru_ixrss, usage.ru_idrss, usage.ru_isrss, usage.ru_majflt, usage.ru_nvcsw, usage.ru_nivcsw);
    }

    /* info per child */
    if ((_child_control) && (_child_control->list_of_children)) {
        int i;
        for (i = 0; i < _child_control->number_of_children; i++) {
            child_t *child = _child_control->list_of_children[i];
            printf ("child-%d: %d runs, %d active\n", i, child->number_of_dead, child->pid != 0);
        }
    }
}

/**
   Clean internal memory
*/
void _clean_child_control (void)
{
    VERBOSE (child_control, 2, printf ("cleaning child control structure\n"));

    if (_child_control) {
        int i;
        int is_child_in_list = 0;
        for (i = 0; i < _child_control->number_of_children; i++)
            if (_child_control->list_of_children[i]) {
                is_child_in_list |= (_child_control->list_of_children[i] == _child);
                free (_child_control->list_of_children[i]);
            }
        if (!is_child_in_list){
            free (_child);
            _child = NULL;
        }

        if (_child_control->list_of_children)
            free (_child_control->list_of_children);

        free (_child_control);

        _child_control = NULL;
    }
}

void kill_all_children (int cause, int sig)
{

    /* check that children control has been properly initialized */
    assert (_child_control && _child_control->father_pid);

    /* check if function has already been called */
    if (_child_control->killing) return;
    _child_control->killing = 1;

    /* display log message */
    if (cause)
        VERBOSE (child_control, 0, printf ("process %d wants to kill all children; cause %d (%d)\n", getpid (), cause, _child_control->father_pid));
    else
        VERBOSE (child_control, 1, printf ("process %d wants to kill all children; cause %d (%d)\n", getpid (), cause, _child_control->father_pid));

    /* send signal to children */
    int i;
    for (i = 0; i < _child_control->number_of_children; i++) {
        child_t *child = _child_control->list_of_children[i];
        if ((child) && (child->pid)) {
            VERBOSE (child_control, 2, printf ("killing pid %d (sig %d)\n", child->pid, sig));
            ptrace (PTRACE_DETACH, child->pid, NULL, NULL);
            kill (child->pid, sig);
        }
    }

    /* for normal exit, display usage */
    if ((sig == SIGTERM) && (_child_control->father_pid == getpid ())) {
        VERBOSE  (child_control, 2, printf ("waiting for children...\n"));
        sleep (1);

        /* get resource usage */
        VERBOSE (child_control, 1, get_resource_usage ());
    }

    _clean_child_control ();
}

/**
   Kill all children by SIGTERM, cause is 0. This function is always used at exit.
*/
void _kill_all_children (void)
{
    /* check that calling process is only the father */
    assert (_child_control && _child_control->father_pid);
    if ((_child_control->father_pid == getpid ()) && (_child_control->number_of_children)) {
        VERBOSE (child_control, 2, printf ("kill all children from exit call\n"));
        kill_all_children (0, SIGTERM);
    }

    _clean_child_control ();
}

/**
   Signal handler

   @param s receive signal
*/
void _signal_handler (int signal)
{

    /* check that children control has been properly initialized */
    assert (_child_control && _child_control->father_pid);
    if (_child_control->killing) return;

    if ((_child_control->father_pid != getpid ()) && (_child))
        VERBOSE (child_control, 1, printf ("signal_handler (%d - %d - %d)\n", signal, getpid (), _child->id));
    else
        VERBOSE (child_control, 1, printf ("signal_handler (%d - %d)\n", signal, getpid ()));

    /* kill all if signal is received by the father */
    if (_child_control->father_pid == getpid ()) {
        VERBOSE (child_control, 2, printf ("kill all children\n"));
        kill_all_children (0, SIGTERM);
    } else {
        /* for child */
        if ((_child) && (_child->term))
            _child->term (_child->data);
    }

    _clean_child_control ();

    _exit (signal);
}

/**
   Wait child signal

   @return 0 if a child dies, -1 on waitpid error, 1 if nobody dies
*/
int _wait_child (void)
{
    /* check that children control has been properly initialized */
    assert (_child_control && _child_control->father_pid);

    /* get pid */
    int status;
    pid_t child_pid = waitpid (-1, &status, 0);
    VERBOSE (child_control, 2, printf ("waitpid: pid=%d status=%d\n", child_pid, status));
    if (child_pid < 0) {
        return -1;
    }

    /* look for pid in children list */
    int id, child_id = -1;
    for (id = 0; id < _child_control->number_of_children; id++) {
        child_t *child = _child_control->list_of_children[id];
        if (child->pid == child_pid) {
            child_id = child->id;
            child->pid = 0;
            child->number_of_dead++; 
            break;
        }
    }
    if  (child_id == -1) {
        VERBOSE (child_control, 2, printf ("no id for pid %d\n", child_pid));
        return 1;
    }

    VERBOSE (child_control, 1, printf ("child %d is dead, pid %d\n", child_id, child_pid));
    if ((WIFSTOPPED (status)) && (WSTOPSIG (status) == SIGTRAP)) {
        VERBOSE (child_control, 2, printf ("trapped pid %d (child %d)\n", child_pid, child_id));
        ptrace(PTRACE_CONT, child_pid, NULL, SIGABRT);

    } else if (WIFSTOPPED (status)) {
        VERBOSE (child_control, 2, printf ("pid %d (child %d) stopped by signal %d\n", child_pid, child_id, WIFSTOPPED (status)));
        if (1 || (WIFSTOPPED(status) == SIGHUP)) {
            ptrace (PTRACE_DETACH, child_pid, NULL, NULL);
            VERBOSE (child_control, 1, printf ("detached\n"));
            kill (child_pid, SIGCONT);
            VERBOSE (child_control, 1,  printf ("contine\n"));
            kill (child_pid, SIGKILL);
            return 0;
        }

    } else if (WIFEXITED (status)) {
        int rc = WEXITSTATUS (status);
        VERBOSE (child_control, 2, printf ("pid %d (child %d) exited rc=%d\n", child_pid, child_id, rc));

    } else if (WIFSIGNALED (status)) {
        int sig = WTERMSIG(status);
        VERBOSE (child_control, 2, printf ("pid %d (child %d) killed sig=%d\n", child_pid, child_id, sig));

        if (sig != SIGABRT) {
            child_t *child = _child_control->list_of_children[id];
            if (child->term)
                child->term (child->data);
        }

    } else {

        VERBOSE (child_control, 2, printf ("pid %d (child %d) exited/killed status 0x%X\n", child_pid, child_id, status));

    }

    return 0;
}

/**
   Start a child process

   @param child structure containing all needed informations

   @return same strucuture with pid field filled with the subprocess pid
*/
child_t *_start_child (child_t *child)
{
    /* check that children control has been properly initialized */
    assert (_child_control && _child_control->father_pid && child);

    /* fork and start child process */
    if (child->preinit)
        child->preinit (child->data);
    child->pid = fork ();
    switch ((int)child->pid) {
    case -1: /* Error */
        VERBOSE (child_control, 0, printf ("error fork (%d), %s\n", errno, strerror (errno)));
        kill_all_children (-1, SIGTERM);
        _exit (CHILD_CONTROL_RC_ASSERT_END);
        break;
        
    case 0: /* Child */ 
        child->pid = getpid ();
        _child = child;
        if (child->process)
            child->process (child->data);
        _exit (CHILD_CONTROL_RC_NORMAL_END);
        break;
    }
    if (child->postinit)
        child->postinit (child->data);
    VERBOSE (child_control, 1, printf ("running child %d running on pid %d\n", child->id, (int) child->pid));

    return child;
}

pid_t register_child (int child_id, int (*preinit) (void *data), int (*postinit) (void *data), int (*process) (void *data), int (*term) (void *data), void *data)
{

    VERBOSE (child_control, 2, printf ("registering child %d\n", child_id));

    /* allocate child structure */
    child_t *child = (child_t *) calloc (sizeof (child_t), 1);
    assert (child);
    child->data = data;
    child->id = child_id;
    child->postinit = postinit;
    child->process = process;
    child->preinit = preinit;
    child->term = term;

    /* register new child into table */
    int id = _child_control->number_of_children;
    _child_control->list_of_children =
        (child_t **) realloc (_child_control->list_of_children, sizeof (child_t *) * (_child_control->number_of_children + 1));
    assert (_child_control->list_of_children);

    _child_control->list_of_children[id] = NULL;
    _child_control->list_of_children[id] = _start_child (child);
    _child_control->number_of_children++;

    VERBOSE (child_control, 2, printf ("register child %d running on pid %d (idx %d)\n", child->id, (int) child->pid, id));

    return child->pid;
}

/**
   Dispatcher function. This function never returns.

   @param unused

   @return NULL
*/
void *_child_control_dispatcher (__attribute__ ((unused)) void *param)
{
    VERBOSE (child_control, 2, printf ("start child control dispatcher\n"));

    /* wait first child registration */
    while (_child_control->number_of_children == 0)
        usleep (100);

    /* main loop */
    while (1) {

        int rc = _wait_child ();
        VERBOSE (child_control, 2, printf ("wait child return: %d\n", rc));

        if (rc == 0) {
            int id;

            for (id = 0; id < _child_control->number_of_children; id++) {
                child_t *child = _child_control->list_of_children[id];
                if (child->pid == 0) {

                    if ((_child_control->max_number_of_dead > 0) &&
                        (++_child_control->number_of_dead >= _child_control->max_number_of_dead)) {
                        VERBOSE (child_control, 0, printf ("too many dead children (%d), aborting\n", _child_control->number_of_dead));
                        kill_all_children (0, SIGTERM);
                        _exit (CHILD_CONTROL_RC_ASSERT_END);
                    }

                    if (!_child_control->killing){
                        VERBOSE (child_control, 1, printf ("restart child %d (%d already dead)\n",  child->id, _child_control->number_of_dead));
                        _start_child (child);
                    }
                }
            }
        } else if (rc == -1) {
            VERBOSE (child_control, 0, printf ("all children are dead (%d deads, %d total), killing dispatcher\n", _child_control->number_of_dead, _child_control->number_of_children));
            _exit (CHILD_CONTROL_RC_MAX_DEAD_END);
        } else {
            VERBOSE (child_control, 1, printf ("something wrong in dispatcher (%d), sleeping 1ms\n", rc));
            usleep (1000);
        }
    }

    VERBOSE (child_control, 0, printf ("stopping, out of dispatcher\n"));

    pthread_exit (NULL);

    return NULL;
}

/**
   Init function
*/
void __attribute__((constructor)) _init_child_control (void)
{
    char *answer;
    assert (!_child_control);

    /* initialise child control */
    _child_control = (child_control_t *) calloc (sizeof (child_control_t), 1);
    assert (_child_control);
    _child_control->father_pid = getpid ();

    /* get maximun number of dead children */
    answer = getenv ("CHILD_CONTROL_MAX_DEAD_");
    _child_control->max_number_of_dead = (answer == NULL) ? CHILD_CONTROL_MAX_DEAD : atoi (answer);

    /* get verbose level */
    answer = getenv ("CHILD_CONTROL_VERBOSE_");
    if (answer)
        CHANGE_VERBOSE_LEVEL (child_control, atoi (answer));
    VERBOSE (child_control, 2, printf ("father pid: %d\n", _child_control->father_pid));

    /* trap signals */
    signal (SIGTERM, _signal_handler);
    signal (SIGQUIT, _signal_handler);
    signal (SIGINT, _signal_handler);

    /* initialized with default attributes */
    int ret = 0;
    pthread_attr_t tattr;
    ret |= pthread_attr_init (&tattr);
    size_t size = 1024 * 1024;
    ret |= pthread_attr_setstacksize (&tattr, size);
    if (ret)
        VERBOSE (child_control, 0, printf ("error while setting thread attributes: %d\n", ret));

    /* create dispatching thread */
    ret = pthread_create (&_child_control->dispatcher_thread, &tattr, _child_control_dispatcher, NULL);
    if (ret)
        VERBOSE (child_control, 0, printf ("error while creating dispatcher thread: %d\n", ret));
    pthread_attr_destroy (&tattr);

    /* kill all children at exit */
    ret = atexit (_kill_all_children);
    if (ret)
        VERBOSE (child_control, 0, printf ("error while adding kill_all_chidren function in exit queue: %d\n", ret));
}

/* vi:set tabstop=4 expandtab shiftwidth=4: this line set vi mode */
